(* An illustration of the trep library: generic printing, or gshow 
   The function gshow takes a value and a representation of its type
   and returns the string `showing' the value:

    val gshow : 'a trep -> 'a -> string

   The function is extensible to user-defined data types.

   The canonical structures facility is a general mechanism to easily
   program such generic functions.
*)

open Trep
(*
#load "trep.cmo";;
*)

(* Mechanism to make gshow extensible *)
type gshow_t = {gshow: 'a. 'a trep -> 'a -> string option}

let gshowref : gshow_t ref = ref {gshow = fun _ _ -> None} (* default case *)

let gshow : type a. a trep -> a -> string = fun t v ->
  Option.value ~default:"?" @@ (!gshowref).gshow t v

(* Registering an extension of gshow *)
let gshow_extend : gshow_t -> unit = fun {gshow = ext} ->
  let {gshow=gshow_old} = !gshowref in
  gshowref := 
   {gshow = fun t v -> match ext t v with None -> gshow_old t v | r -> r}

(* Registering the initial extension *)
let () = 
  let gshow_init : type a. a trep -> a -> string option = fun t v ->
    match t with
    | Unit   -> Some "()"
    | Int    -> Some (string_of_int v)
    | Bool   -> Some (string_of_bool v)
    | String -> Some v
    | Fun _  -> Some "<fun>"
    | Option t' -> 
        begin match v with
        | Some v' -> (!gshowref).gshow t' v' |> Option.map ((^) "Some ") 
        | _       -> Some "None"
        end
    | Pair (t1,t2) ->
        begin match ((!gshowref).gshow t1 (fst v), (!gshowref).gshow t2 (snd v))
            with
            | (Some x, Some y) -> Some ("(" ^ x ^ ", " ^ y ^ ")")
            | _ -> None
        end
    | List t when v = [] -> Some "[]" (* we can show an empty list always *)
    | List t ->
        List.map ((!gshowref).gshow t) v |> fun l ->
        List.fold_right 
          (fun e l -> match (e,l) with (Some e,Some l) -> Some (e::l) 
                      | _ -> None)
          l (Some []) |> Option.map (fun l -> "[" ^ String.concat "; " l ^ "]")
    | _  -> None
  in gshow_extend {gshow = gshow_init}
;;

(* Tests *)

let _ = gshow (Pair (Int,Bool)) (1,true)
(*
- : string = "(1, true)"
*)

let _ = gshow (List (Pair (Int,Bool))) [(1,true); (2,false)]
(*
- : string = "[(1, true); (2, false)]"
*)

let _ = gshow (List (Pair (Int,Bool))) []
(*
- : string = "[]"
*)

(* Showing a user-defined data type *)

(* A user-defined data type *)
type 'a myres = R of 'a * int | Err of string

(* registering it with the trep library *)
type _ trep += MyRes : 'a trep -> 'a myres trep

let () = 
  let teq : type a b. a trep -> b trep -> (a,b) eq option = fun x y -> 
    match (x,y) with 
    | (MyRes x,MyRes y) ->(match teq x y with Some Refl-> Some Refl | _ -> None)
    | _ -> None
  in teq_extend {teq}

(* Adding a clause to gshow *)

let () = 
  let gshow : type a. a trep -> a -> string option = function MyRes t ->
    (function R (x,i) -> Some ("R " ^ gshow (Pair (t,Int)) (x,i))
             | Err s  -> Some ("Err " ^ s))
    | _ -> fun _ -> None
  in gshow_extend {gshow}

let _ = gshow (Pair (Bool,MyRes (List Int))) (true,R ([1;2;3],4))
(*
- : string = "(true, R ([1; 2; 3], 4))"
*)

;;
